@@ -438,7 +438,7 @@ function(node)
 
 
 setGeneric("xmlSourceFunctions",
-            function(doc, ids = character(), ...) {
+            function(doc, ids = character(), parse = TRUE, ...) {
                   standardGeneric("xmlSourceFunctions")
             })
 
@@ -447,9 +447,9 @@ setMethod("xmlSourceFunctions", "character",
   #
   # evaluate the r:function nodes, or restricted to @id from ids.
   #
-function(doc, ids = character(), ...)
+function(doc, ids = character(), parse = TRUE, ...)
 {
-  xmlSourceFunctions(xmlParse(doc), ids, ...)
+  xmlSourceFunctions(xmlParse(doc), ids, parse = parse, ...)
 })
 
 
@@ -461,7 +461,7 @@ setMethod("xmlSourceFunctions", "XMLInternalDocument",
   #
   # evaluate the r:function nodes, or restricted to @id from ids.
   #
-function(doc, ids = character(), ...)
+function(doc, ids = character(), parse = TRUE, ...)
 {
 
   if(length(ids))
@@ -469,6 +469,9 @@ function(doc, ids = character(), ...)
   else
      nodes = getNodeSet(doc, "//r:function", c(r = "http://www.r-project.org"))
 
+  if(parse == FALSE)
+     return(nodes)
+
   xmlSource(nodes, ...)
 })
 
