@@ -17,8 +17,11 @@ function(libname, pkgname)
 
   # Added by Uwe Ligges.
  if(.Platform$OS.type == "windows"){
+
+     fixPath = if(getRversion() < "2.13.0")  utils::normalizePath else base::normalizePath
+
      temp <- Sys.getenv("PATH")
-     Sys.setenv("PATH" = paste(utils::normalizePath(file.path(libname, pkgname, "libs")), 
+     Sys.setenv("PATH" = paste(fixPath(file.path(libname, pkgname, "libs")), 
                                file.path(Sys.getenv("R_HOME"), "modules", fsep="\\"), temp, sep=";"))
      on.exit(Sys.setenv(PATH = temp))
  }     
